/*
Module : CPUTICKER.CPP
Purpose: Implementation for a high resolution counter class. The class wraps the Pentium-specific 
         time stamp counter available via the RDTSC assembly language instruction. This counter has 
         a resolution in terms of PCLKS (processor clocks) so it can be used for direct instruction 
         timings. Please note that you should be aware of various issues associated with using the
         RDTSC instruction for timing purposes. For further information on the issues please see 
         the MSDN article at http://msdn.microsoft.com/library/default.asp?url=/library/en-us/directx9_c/Game_Timing_and_Multicore_Processors.asp
Created: J.M.McGuiness / 26/3/96
History: PJ Naughter / 16/7/97  
         A number of additions including:
         1. Support for running on Windows NT
         2. now uses the build-in 64 bit data type "__int64"
         3. Improved diagnostic info thanks to the above
         4. Usage of static variables to improve efficiency
         5. Addition of a function which will convert from CPU ticks to seconds
         6. Improved adherence to the MFC coding style and standards
         PJ Naughter / 14/1/99
         1. Fixed a bug discovered by David Green-Seed where he was experiencing 
         access violations when the code was compiled with optimizations turned 
         on a Pentium II. The problem was that (on a PII) the RDTSC instruction 
         touches more registers than expected. This has now been fixed by saved 
         and restoring the EAX and EBX registers around the call to RDTSC.
         2. General code tidy up of const functions, parameters etc.
         PJ Naughter 03/12/99
         1. Fixed a problem in when compiling in VC 6 in release mode.
         PJ Naughter 17/05/00
         1. Measure now returns the value measured.
         2. Code is now UNICODE enabled and build configurations are provided
         PJ Naughter 30/05/01
         1. Minor changes to the GetTickCountAsSeconds function.
         2. Updated the sample app to report the processor frequency
         3. Added a GetCachedCPUFrequency function
         PJ Naughter 23/02/04
         1. Now uses VC 6 SP5's and later support for built in assembler support for
         cpuid and rdtsc. Thanks to Matthew Hurd for reporting this. This does mean
         that the code will only compile on VC 6 SP5 or later now.
         PJ Naughter 24/12/06
         1. Updated copyright details
         2. Removed derivation from CObject as it was not really required
         3. Updated documentation to use the same style as my web site
         4. Addition of a CCMC_EXT_CLASS preprocessor macro to allow the classes to be 
         more easily added to an extension dll
         5. Reviewed all TRACE statements for correctness
         6. Code now uses newer C++ style casts instead of C style casts.
         7. Removed code which disables interrupts from CCPUTicker::Measure
         8. Code now uses "cmath" instead of math.h"
         9. Code now assumes that RDTSC is always available. This is no longer
         a big issue since there are not that many 386/486 machines around!
         10. Updated the code to clean compile on VC 2005
         11. Updated the code to support x64 bit Windows
         12. Removed the member variable m_TickCount from the class
         13. Changed the default interval for GetCPUFrequency to 100 milliseconds
         from 1000 milliseconds.
         14. Removed now unnecessary CCPUTicker constructor

Copyright (c) 2003 - 2006 by J.M.McGuiness (www.hussar.demon.co.uk)
                             PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


///////////////////////////////// Includes  ///////////////////////////////////

#include "stdafx.h"
#ifndef _CMATH_
#pragma message("To avoid this message please put cmath into your PCH (normally stdafx.h)")
#include <cmath>
#endif
#include "CPUTicker.h"
#if (_MSC_VER >= 1400)
#ifndef __INTRIN_H_
#pragma message("To avoid this message please put intrin.h into your PCH (normally stdafx.h)")
#include <intrin.h>
#endif
#endif


///////////////////////////////// Macros / Defines ////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

double CCPUTicker::m_dAverageDeviation = 0;
double CCPUTicker::m_dCPUFrequency = 0;
BOOL CCPUTicker::m_bFrequencyCalculated = FALSE;


//////////////////////////////// Implementation ///////////////////////////////

#if defined _M_IX86
  #if (_MSC_VER < 1400)
    #pragma optimize("", off)
  #endif
#endif
unsigned __int64 CCPUTicker::Measure()
{
#if defined _M_IX86
  #if (_MSC_VER >= 1400)
    return __rdtsc();
  #else
		volatile ULARGE_INTEGER ts;
		_asm
		{
      cpuid
      rdtsc
			mov ts.HighPart,edx
			mov ts.LowPart,eax
		}
    return ts.QuadPart;
  #endif
#elif defined _M_X64
  //Note we only support x64 on VC 2005 or later
  return __rdtsc();
#else
  #pragma message("The CCPUTicker class is currently not supported on this processor architecture")
  return 0;
#endif
}
#if defined _M_IX86
  #if (_MSC_VER < 1400)
    #pragma optimize("", on)
  #endif
#endif

//The following function will work out the processor clock frequency to a
//specified accuracy determined by the target average deviation required.
//Note that the worst average deviation of the result is less than 5MHz for
//a mean frequency of 90MHz. So basically the target average deviation is
//supplied only if you want a more accurate result, it won't let you get a
//worse one. (Units are Hz.)
//
//(The average deviation is a better and more robust measure than it's cousin
//the standard deviation of a quantity. The item determined by each is
//essentially similar. See "Numerical Recipies", W.Press et al for more
//details.)
//
//This function will run for a maximum of dwMapsLoops*dwInterval milliseconds 
//before giving up on trying to improve the average deviation, with the average 
//deviation actually achieved replacing the supplied target value. Use 
//"dwMaxloops" to change this. 
#pragma optimize("",off)
BOOL CCPUTicker::GetCPUFrequency(double& CPUFrequency, double& TargetAverageDeviation, DWORD dwInterval, DWORD dwMaxLoops)
{
	//double ave_dev,tmp = 0;
	
	//Work out how long to wait for in QPC units
	LARGE_INTEGER nWaitPeriod;
	if (!QueryPerformanceFrequency(&nWaitPeriod))
		return FALSE;
	nWaitPeriod.QuadPart *= dwInterval;
	nWaitPeriod.QuadPart /= 1000;

	//Start of tight timed loop.
	LARGE_INTEGER nQPCGoal;
	QueryPerformanceCounter(&nQPCGoal);
	nQPCGoal.QuadPart += nWaitPeriod.QuadPart;
  CCPUTicker counter;
	unsigned __int64 nStartCounter = counter.Measure();
	LARGE_INTEGER nQPCValue;
	do
	{
		QueryPerformanceCounter(&nQPCValue);
	} 
	while (nQPCValue.QuadPart < nQPCGoal.QuadPart);
	unsigned __int64 nFinishCounter = counter.Measure();
	//End of tight timed loop.

	DWORD dwCounter = 0;
	double dFrequencyDifference = 0;
	double dAverageDeviation = 0;
#if (_MSC_VER >= 1300)
	double dAverageFrequency = static_cast<double>((nFinishCounter - nStartCounter) / dwInterval * 1000);
#else
  double dAverageFrequency = static_cast<double>(static_cast<__int64>((nFinishCounter - nStartCounter) / dwInterval * 1000));
#endif
	do
	{
		//Start of tight timed loop.
		QueryPerformanceCounter(&nQPCGoal);
		nQPCGoal.QuadPart += nWaitPeriod.QuadPart;
		nStartCounter = counter.Measure();
		do
		{
			QueryPerformanceCounter(&nQPCValue);
		} 
		while (nQPCValue.QuadPart < nQPCGoal.QuadPart);
		nFinishCounter = counter.Measure();
		//End of tight timed loop.

		//Average of the old frequency plus the new.
  #if (_MSC_VER >= 1300)
  	double dCurrentFrequency = static_cast<double>((nFinishCounter - nStartCounter) / dwInterval * 1000);
  #else
    double dCurrentFrequency = static_cast<double>(static_cast<__int64>((nFinishCounter - nStartCounter) / dwInterval * 1000));
  #endif
		dAverageFrequency = (dCurrentFrequency + dAverageFrequency)/2;

		//Work out the current average deviation of the frequency.
		dFrequencyDifference += fabs(dCurrentFrequency - dAverageFrequency);
		dAverageDeviation = dFrequencyDifference/(++dwCounter);
	} 
	while ((dAverageDeviation > TargetAverageDeviation) && (dwCounter < dwMaxLoops));
	TargetAverageDeviation = dAverageDeviation;
  CPUFrequency = dAverageFrequency;
	TRACE(_T("CCPUTicker::GetCPUFrequency, Estimated the processor clock frequency=%gHz, dev.=%gHz.\n"), CPUFrequency, dAverageDeviation);

  //Cache the results in the member variables
  m_bFrequencyCalculated = TRUE;
  m_dCPUFrequency = CPUFrequency;
  m_dAverageDeviation = dAverageDeviation;

	return TRUE;
}
#pragma optimize("",on)


BOOL CCPUTicker::GetCachedCPUFrequency(double& CPUFrequency, double& AverageDeviation)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;
  
  if (m_bFrequencyCalculated)
  {
    CPUFrequency = m_dCPUFrequency;
    AverageDeviation = m_dAverageDeviation;
    bSuccess = TRUE;
  }
  else
    bSuccess = CCPUTicker::GetCPUFrequency(CPUFrequency, AverageDeviation);
    
  return bSuccess;
}

double CCPUTicker::GetTickCountAsSeconds(unsigned __int64 nValue)
{
  if (!m_bFrequencyCalculated)
  {
    BOOL bSuccess = CCPUTicker::GetCachedCPUFrequency(m_dCPUFrequency, m_dAverageDeviation);
    if (!bSuccess)
      return 0;
  }

#if (_MSC_VER >= 1300)
  return (static_cast<double>(nValue)) / m_dCPUFrequency;
#else
  return (static_cast<double>(static_cast<__int64>(nValue))) / m_dCPUFrequency;
#endif
}
